'use strict';

const Module = require('module');
const upath = require('upath');
const chalk = require('chalk');
const consola = require('consola');
const defu = require('defu');
const prettyBytes = require('pretty-bytes');
const jiti = require('jiti');
const mkdirp = require('mkdirp');
const fsp = require('fs/promises');
const util = require('util');
const _rimraf = require('rimraf');
const rollup = require('rollup');
const commonjs = require('@rollup/plugin-commonjs');
const pluginNodeResolve = require('@rollup/plugin-node-resolve');
const alias = require('@rollup/plugin-alias');
const esbuild = require('rollup-plugin-esbuild');
const dts = require('rollup-plugin-dts');
const untyped = require('untyped');
const untypedPlugin = require('untyped/loader/babel');
const scule = require('scule');
const mkdist = require('mkdist');

function _interopDefaultLegacy (e) { return e && typeof e === 'object' && 'default' in e ? e['default'] : e; }

const Module__default = /*#__PURE__*/_interopDefaultLegacy(Module);
const chalk__default = /*#__PURE__*/_interopDefaultLegacy(chalk);
const consola__default = /*#__PURE__*/_interopDefaultLegacy(consola);
const defu__default = /*#__PURE__*/_interopDefaultLegacy(defu);
const prettyBytes__default = /*#__PURE__*/_interopDefaultLegacy(prettyBytes);
const jiti__default = /*#__PURE__*/_interopDefaultLegacy(jiti);
const mkdirp__default = /*#__PURE__*/_interopDefaultLegacy(mkdirp);
const fsp__default = /*#__PURE__*/_interopDefaultLegacy(fsp);
const _rimraf__default = /*#__PURE__*/_interopDefaultLegacy(_rimraf);
const commonjs__default = /*#__PURE__*/_interopDefaultLegacy(commonjs);
const alias__default = /*#__PURE__*/_interopDefaultLegacy(alias);
const esbuild__default = /*#__PURE__*/_interopDefaultLegacy(esbuild);
const dts__default = /*#__PURE__*/_interopDefaultLegacy(dts);
const untypedPlugin__default = /*#__PURE__*/_interopDefaultLegacy(untypedPlugin);

async function ensuredir(path) {
  await mkdirp__default(upath.dirname(path));
}
async function symlink(from, to, force = true) {
  await ensuredir(to);
  if (force) {
    await fsp__default.unlink(to).catch(() => {
    });
  }
  await fsp__default.symlink(from, to, "junction");
}
function dumpObject(obj) {
  return "{ " + Object.keys(obj).map((key) => `${key}: ${JSON.stringify(obj[key])}`).join(", ") + " }";
}
function getpkg(id = "") {
  const s = id.split("/");
  return s[0][0] === "@" ? `${s[0]}/${s[1]}` : s[0];
}
const rimraf = util.promisify(_rimraf__default);
async function rmdir(dir) {
  await fsp__default.unlink(dir).catch(() => {
  });
  await rimraf(dir);
}

function validateDependencies(ctx) {
  const usedDependencies = new Set();
  const unusedDependencies = new Set(Object.keys(ctx.pkg.dependencies || {}));
  const implicitDependnecies = new Set();
  for (const id of ctx.usedImports) {
    unusedDependencies.delete(id);
    usedDependencies.add(id);
  }
  if (Array.isArray(ctx.dependencies)) {
    for (const id of ctx.dependencies) {
      unusedDependencies.delete(id);
    }
  }
  for (const id of usedDependencies) {
    if (!ctx.externals.includes(id) && !id.startsWith("chunks/") && !ctx.dependencies.includes(getpkg(id))) {
      implicitDependnecies.add(id);
    }
  }
  if (unusedDependencies.size) {
    consola__default.warn("Potential unused dependencies found:", Array.from(unusedDependencies).map((id) => chalk__default.cyan(id)).join(", "));
  }
  if (implicitDependnecies.size && !ctx.inlineDependencies) {
    consola__default.warn("Potential implicit dependencies found:", Array.from(implicitDependnecies).map((id) => chalk__default.cyan(id)).join(", "));
  }
}

async function rollupBuild(ctx) {
  if (ctx.stub) {
    for (const entry of ctx.entries.filter((entry2) => entry2.builder === "rollup")) {
      const output = upath.resolve(ctx.rootDir, ctx.outDir, entry.name);
      await fsp.writeFile(output + ".cjs", `module.exports = require('jiti')()('${entry.input}')`);
      await fsp.writeFile(output + ".mjs", `import jiti from 'jiti';
export default jiti()('${entry.input}');`);
      await fsp.writeFile(output + ".d.ts", `export * from '${entry.input}'`);
    }
    return;
  }
  const rollupOptions = getRollupOptions(ctx);
  if (!Object.keys(rollupOptions.input).length) {
    return;
  }
  const buildResult = await rollup.rollup(rollupOptions);
  const allOutputOptions = rollupOptions.output;
  for (const outputOptions of allOutputOptions) {
    const { output } = await buildResult.write(outputOptions);
    for (const entry of output.filter((e) => e.type === "chunk")) {
      for (const id of entry.imports) {
        ctx.usedImports.add(id);
      }
      if (entry.isEntry) {
        ctx.buildEntries.push({
          path: upath.relative(ctx.rootDir, upath.resolve(outputOptions.dir, entry.fileName)),
          bytes: entry.code.length * 4,
          exports: entry.exports
        });
      }
    }
  }
  if (ctx.declaration) {
    rollupOptions.plugins = rollupOptions.plugins || [];
    rollupOptions.plugins.push(dts__default({
      respectExternal: true
    }));
    const typesBuild = await rollup.rollup(rollupOptions);
    await typesBuild.write({
      dir: upath.resolve(ctx.rootDir, ctx.outDir),
      format: "esm"
    });
  }
}
function getRollupOptions(ctx) {
  const extensions = [".ts", ".mjs", ".cjs", ".js", ".json"];
  return {
    context: ctx.rootDir,
    input: Object.fromEntries(ctx.entries.filter((entry) => entry.builder === "rollup").map((entry) => [entry.name, upath.resolve(ctx.rootDir, entry.input)])),
    output: [
      {
        dir: upath.resolve(ctx.rootDir, ctx.outDir),
        entryFileNames: "[name].cjs",
        chunkFileNames: "chunks/[name].cjs",
        format: "cjs",
        exports: "auto",
        preferConst: true,
        externalLiveBindings: false,
        freeze: false
      },
      {
        dir: upath.resolve(ctx.rootDir, ctx.outDir),
        entryFileNames: "[name].mjs",
        chunkFileNames: "chunks/[name].mjs",
        format: "esm",
        exports: "auto",
        preferConst: true,
        externalLiveBindings: false,
        freeze: false
      }
    ],
    external(id) {
      const pkg = getpkg(id);
      const isExplicitExternal = ctx.externals.includes(pkg);
      if (isExplicitExternal) {
        return true;
      }
      if (ctx.inlineDependencies || id[0] === "." || id[0] === "/" || id.match(/src[\\/]/) || id.startsWith(ctx.pkg.name)) {
        return false;
      }
      if (!isExplicitExternal) {
        consola__default.warn(`Inlining implicit external ${id}`);
      }
      return isExplicitExternal;
    },
    onwarn(warning, rollupWarn) {
      if (!warning.code || !["CIRCULAR_DEPENDENCY"].includes(warning.code)) {
        rollupWarn(warning);
      }
    },
    plugins: [
      alias__default({
        [ctx.pkg.name]: ctx.rootDir
      }),
      pluginNodeResolve.nodeResolve({
        extensions,
        preferBuiltins: true
      }),
      {
        name: "json",
        transform(json, id) {
          if (!id || id[0] === "\0" || !id.endsWith(".json")) {
            return null;
          }
          return {
            code: "module.exports = " + json,
            map: null
          };
        }
      },
      esbuild__default({
        target: "es2019"
      }),
      commonjs__default({
        extensions,
        ignore: [
          ...ctx.externals
        ]
      })
    ]
  };
}

async function typesBuild(ctx) {
  for (const entry of ctx.entries.filter((entry2) => entry2.builder === "untyped")) {
    const _require = jiti__default(ctx.rootDir, {
      interopDefault: true,
      transformOptions: {
        babel: {
          plugins: [
            untypedPlugin__default
          ]
        }
      }
    });
    const distDir = entry.outDir;
    const srcConfig = _require(upath.resolve(ctx.rootDir, entry.input));
    const defaults = entry.defaults || {};
    const schema = untyped.resolveSchema(srcConfig, defaults);
    await fsp.writeFile(upath.resolve(distDir, `${entry.name}.md`), untyped.generateMarkdown(schema));
    await fsp.writeFile(upath.resolve(distDir, `${entry.name}.schema.json`), JSON.stringify(schema, null, 2));
    await fsp.writeFile(upath.resolve(distDir, `${entry.name}.defaults.json`), JSON.stringify(defaults, null, 2));
    if (entry.declaration) {
      await fsp.writeFile(upath.resolve(distDir, `${entry.name}.d.ts`), "export " + untyped.generateTypes(schema, scule.pascalCase(entry.name + "-schema")));
    }
  }
}

async function mkdistBuild(ctx) {
  for (const entry of ctx.entries.filter((e) => e.builder === "mkdist")) {
    const distDir = entry.outDir;
    if (ctx.stub) {
      await rmdir(distDir);
      await symlink(entry.input, distDir);
    } else {
      const { writtenFiles } = await mkdist.mkdist({
        rootDir: ctx.rootDir,
        srcDir: entry.input,
        distDir,
        format: entry.format,
        cleanDist: false,
        declaration: entry.declaration,
        ext: entry.ext
      });
      ctx.buildEntries.push({
        path: distDir,
        chunks: [`${writtenFiles.length} files`]
      });
    }
  }
}

async function build(rootDir, stub) {
  rootDir = upath.resolve(process.cwd(), rootDir || ".");
  const _require = jiti__default(rootDir);
  let buildConfigFile;
  try {
    buildConfigFile = _require.resolve("./build.config");
  } catch (e) {
  }
  const buildConfig = buildConfigFile ? _require("./build.config").default : {};
  const pkg = _require("./package.json");
  const ctx = defu__default(buildConfig, {
    pkg,
    rootDir,
    entries: [],
    dependencies: [],
    devDependencies: [],
    externals: [...Module__default.builtinModules],
    outDir: "dist",
    genDir: ".gen",
    untyped: void 0,
    declaration: void 0,
    inlineDependencies: false,
    clean: true,
    stub,
    buildEntries: [],
    usedImports: new Set()
  });
  ctx.entries = ctx.entries.map((entry) => typeof entry === "string" ? { input: entry } : entry);
  for (const entry of ctx.entries) {
    if (typeof entry.name !== "string") {
      entry.name = upath.basename(entry.input);
    }
    if (!entry.input) {
      throw new Error("Missing entry input: " + dumpObject(entry));
    }
    if (!entry.builder) {
      entry.builder = entry.input.endsWith("/") ? "mkdist" : "rollup";
    }
    if (ctx.declaration !== void 0 && entry.declaration === void 0) {
      entry.declaration = ctx.declaration;
    }
    entry.input = upath.resolve(ctx.rootDir, entry.input);
    entry.outDir = upath.resolve(ctx.rootDir, entry.outDir || ctx.outDir);
  }
  ctx.dependencies = Object.keys(pkg.dependencies || {});
  ctx.devDependencies = Object.keys(pkg.devDependencies || {});
  ctx.externals.push(...ctx.dependencies);
  consola__default.info(chalk__default.cyan(`${ctx.stub ? "Stubbing" : "Building"} ${pkg.name}`));
  if (process.env.DEBUG) {
    consola__default.info(`${chalk__default.bold("Root dir:")} ${ctx.rootDir}
  ${chalk__default.bold("Entries:")}
  ${ctx.entries.map((entry) => "  " + dumpObject(entry)).join("\n  ")}
`);
  }
  if (ctx.clean) {
    for (const dir of new Set(ctx.entries.map((e) => e.outDir).sort())) {
      await rmdir(dir);
      await mkdirp__default(dir);
    }
  }
  await typesBuild(ctx);
  await mkdistBuild(ctx);
  await rollupBuild(ctx);
  if (ctx.stub) {
    return;
  }
  consola__default.success(chalk__default.green("Build succeed for " + pkg.name));
  for (const entry of ctx.buildEntries) {
    consola__default.log(`  ${chalk__default.bold(entry.path)} (` + [
      entry.bytes && `size: ${chalk__default.cyan(prettyBytes__default(entry.bytes))}`,
      entry.exports && `exports: ${chalk__default.gray(entry.exports.join(", "))}`,
      entry.chunks && `chunks: ${chalk__default.gray(entry.chunks.join(", "))}`
    ].filter(Boolean).join(", ") + ")");
  }
  validateDependencies(ctx);
  consola__default.log("");
}

exports.build = build;
